package ex2;

import java.util.ArrayList;
import java.util.List;

/**
 * An observable subclass of {@link java.util.ArrayList ArrayList}, which
 * notifies other objects (ListObserver) when it is modified. 
 * Notifications are made using ListEvent objects.
 */
public class MyArrayList <E> extends ArrayList<E> implements Observable{

	/* 
	   ArrayList implements Serializable, and it is strongly recommended for 
	   a serializable class to declare its own serialVersionUID
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * The objects that monitor this list aka observers that the list will 
	 * notify when an event occurs.
	 */
	protected List <ListObserver> observers = new ArrayList<ListObserver>();

	/**
	 * Registers an observer that will receive notifications about the 
	 * list's events.
	 * @param o the observer
	 */
	@Override
	public void addObserver(ListObserver o) {
		observers.add(o);
	}

	/**
	 * Removes an observer. This object will no longer be notified about the
	 * list's events.
	 */
	@Override
	public void removeObserver(ListObserver o) {
		observers.remove(o);		
	}
	
	/**
	 * Appends an element to the list and notifies all its observers about this
	 * event.
	 * @param e element to be appended to this list
	 * @return true (as specified by 
	 * {@link java.util.Collection#add Collection.add(E)})
	 */
	@Override
	public boolean add(E e){

		long start = System.nanoTime();
		boolean ret = super.add(e);
		long duration =  System.nanoTime() - start;

		ListEvent event= createEvent(EventType.ADD, e, duration);

		notifyObservers(event);
		return ret;
	}
	
	/**
	 * Replaces the element at the specified position in this list with the
	 * specified element and notifies all the list's observers about this 
	 * event.
	 * @param index index of the element to replace
	 * @param element element to be stored at the specified position
	 * @return the element previously at the specified position
	 */
	@Override
	public E set(int index, E element){
		
		long start = System.nanoTime();
		E ret = super.set(index, element);
		long duration =  System.nanoTime() - start;

		ListEvent event= createEvent(EventType.SET, element, duration);

		notifyObservers(event);

		return ret;
	}
	
	/**
	 * Removes the element at the specified position in this list 
	 * {@link java.util.ArrayList#remove ArrayList.remove(int)}
	 *  and notifies
	 * all the list's observers about this event.
	 * @param index the index of the element to be removed
	 * @return the element that was removed from the list
	 */
	@Override
	public E remove(int index){
		
		long start = System.nanoTime();
		E ret = super.remove(index);
		long duration =  System.nanoTime() - start;

		ListEvent event= createEvent(EventType.REMOVE, ret, duration);

		notifyObservers(event);

		return ret;
	}

	private ListEvent createEvent(final EventType type, 
									final E elem, final long duration){
		
		/*
		 * For simplicity and because it's just a lab exercise to try out inner
		 * classes syntax, we override the abstract class ListEvent in the same
		 * way for all events. 
		 * 
		 * Observe:
		 *  - how to pass parameters to the extended class's constructor
		 *  - how to access the outer class's instance
		 *  - how to access the method's parameters
		 */
		
		return  new ListEvent(type){

			@Override
			@SuppressWarnings("rawtypes")
			public MyArrayList getList() {
				return MyArrayList.this;				
			}

			@Override
			public Object getElement() {
				return elem;
			}

			@Override
			public long getDuration() {
				return duration;
			}			
		};		
	}


	/**
	 * Notifies all the ListObserver objects registered by this list.
	 * @param e the event that occurred on the list
	 */
	protected void notifyObservers(ListEvent e){
		for(ListObserver o : observers){
			o.notifyEvent(e);
		}
	}
	
	@Override
	public String toString(){
		String s = "[";
		String separator = ", ";
		for (int i = 0; i < size()-1; i++){
			s += get(i) + separator;
		}
		s += get(size()-1)+ "]";
		return s;
	}


}
